package gs.test.v0.definition

import cats.Show
import cats.data.EitherT

/** Each instance of this class indicates the _definition_ of some test.
  *
  * @param name
  *   The display name of the test. Not considered to be unique.
  * @param permanentId
  *   The [[PermanentId]] for this test.
  * @param documentation
  *   The documentation for this test.
  * @param tags
  *   The list of [[Tag]] applicable to this test.
  * @param markers
  *   The list of [[Marker]] applicable to this test.
  * @param iterations
  *   The number of iterations of this test to run.
  * @param unitOfWork
  *   The function that the test evaluates.
  */
final class TestDefinition[F[_]](
  val name: TestDefinition.Name,
  val permanentId: PermanentId,
  val documentation: Option[String],
  val tags: List[Tag],
  val markers: List[Marker],
  val iterations: TestIterations,
  val unitOfWork: EitherT[F, TestFailure, Unit]
)

object TestDefinition:

  /** Opaque type representing names that may be assigned to [[Test]].
    */
  opaque type Name = String

  object Name:

    /** Instantiate a new [[Test.Name]]. This name is not unique, has no
      * constraints, and only exists for display purposes.
      *
      * @param name
      *   The candidate string.
      * @return
      *   The new [[Test.Name]] instance.
      */
    def apply(name: String): Name = name

    given CanEqual[Name, Name] = CanEqual.derived

    given Show[Name] = name => name

  end Name

end TestDefinition
