package gs.test.v0.definition

import gs.test.v0.definition.pos.SourcePosition

/** Base trait for all failures recognized by gs-test.
  */
sealed trait TestFailure

object TestFailure:

  /** Returned when assertions in this library fail. Assertions understand how
    * to populate these values.
    *
    * @param assertionName
    *   The name of the assertion.
    * @param inputs
    *   The names and calculated types of each input to the assertion.
    * @param message
    *   The message produced by the assertion.
    * @param pos
    *   The [[SourcePosition]] where the assertion failed.
    */
  case class AssertionFailed(
    assertionName: String,
    inputs: Map[String, String],
    message: String,
    pos: SourcePosition
  ) extends TestFailure

  /** Return when a test explicitly calls `fail("...")` or some variant thereof.
    *
    * @param message
    *   The failure message provided by the test author.
    * @param pos
    *   The [[SourcePosition]] where failure was requested.
    */
  case class TestRequestedFailure(
    message: String,
    pos: SourcePosition
  ) extends TestFailure

  /** Used when the test fails due to an exception.
    *
    * @param cause
    *   The underlying cause of failure.
    */
  case class ExceptionThrown(
    cause: Throwable
  ) extends TestFailure

end TestFailure
