package gs.test.v0.definition

import cats.data.EitherT
import cats.effect.Async
import cats.syntax.all.*
import java.util.concurrent.ConcurrentHashMap
import scala.collection.mutable.ListBuffer
import scala.jdk.CollectionConverters.*

/** Base class for defining groups of related tests. Users should extend this
  * class to define their tests.
  *
  * ## Example
  *
  * {{{
  * import gs.test.v0.definition.*
  *
  * final class MyTestGroup extends TestGroup.IO:
  *   override def name: String = "My Test Group"
  *   override def tags: List[Tag] = List(tag"ex1", tag"ex2")
  *   override def documentation: Option[String] = Some("docs")
  *
  *   test(pid"PermanentId-001", "Example Test Name").pure {
  *     Right(())
  *   }
  * }}}
  */
abstract class TestGroup[F[_]: Async]:
  /** @return
    *   The display name for this group.
    */
  def name: String

  /** @return
    *   List of [[Tag]] that apply to all tests within this group.
    */
  def tags: List[Tag] = List.empty

  /** @return
    *   List of all [[Marker]] that apply to all tests within this group.
    */
  def markers: List[Marker] = List.empty

  /** @return
    *   The documentation for this group.
    */
  def documentation: Option[String] = None

  private var beforeGroupValue: Option[F[Unit]]    = None
  private var afterGroupValue: Option[F[Unit]]     = None
  private var beforeEachTestValue: Option[F[Unit]] = None
  private var afterEachTestValue: Option[F[Unit]]  = None

  private val registry: TestGroup.Registry[F] = new TestGroup.Registry[F]

  /** Compile the contents of this [[TestGroup]] for delivery to the engine.
    *
    * @return
    *   The immutable, compiled form of this [[TestGroup]].
    */
  def compile(): TestGroupDefinition[F] =
    new TestGroupDefinition[F](
      name = TestGroupDefinition.Name(name),
      documentation = documentation,
      testTags = tags,
      testMarkers = markers,
      beforeGroup = beforeGroupValue,
      afterGroup = afterGroupValue,
      beforeEachTest = beforeEachTestValue,
      afterEachTest = afterEachTestValue,
      tests = registry.toList()
    )

  /** Provide an effect that must run before any of the tests within this group
    * are executed.
    *
    * @param f
    *   The effect to run.
    */
  protected def beforeGroup(f: => F[Unit]): Unit =
    beforeGroupValue = Some(f)
    ()

  /** Provide an effect that must run after all tests within this group have
    * finished execution.
    *
    * @param f
    *   The effect to run.
    */
  protected def afterGroup(f: => F[Unit]): Unit =
    afterGroupValue = Some(f)
    ()

  /** Provide an effect that must run before each test within this group.
    *
    * @param f
    *   The effect to run.
    */
  protected def beforeEachTest(f: => F[Unit]): Unit =
    beforeEachTestValue = Some(f)
    ()

  /** Provide an effect that must run after each test within this group.
    *
    * @param f
    *   The effect to run.
    */
  protected def afterEachTest(f: => F[Unit]): Unit =
    afterEachTestValue = Some(f)
    ()

  /** Define a new test.
    *
    * ## Required Information
    *
    * All tests require 3 things, at minimum:
    *
    *   - [[PermanentId]]
    *   - Display Name
    *   - Unit of Work (the code to execute)
    *
    * The [[PermanentId]] must be unique within the [[TestSuite]].
    *
    * ## Default Values
    *
    * Tests iterate 1 time by default. Tags and Markers are inherited from the
    * parent group. If this group contains tag "foo", any test within this group
    * will also get tag "foo".
    *
    * @param permanentId
    *   The [[PermanentId]] for this test.
    * @param name
    *   The display name for this test.
    * @return
    *   A builder, to help complete test definition.
    */
  protected def test(
    permanentId: PermanentId,
    name: String
  ): TestGroup.TestBuilder[F] =
    new TestGroup.TestBuilder[F](
      registry = registry,
      name = TestDefinition.Name(name),
      permanentId = permanentId,
      tags = ListBuffer(tags*),
      markers = ListBuffer(markers*)
    )

object TestGroup:

  /** Specialization of [[TestGroup]] for `cats.effect.IO`, the typical use
    * case.
    */
  abstract class IO extends TestGroup[cats.effect.IO]

  /** Builder to assist with defining tests.
    *
    * @param registry
    *   Registry instance internal to a [[TestGroup]] for recording completed
    *   definitions.
    * @param name
    *   The name of the test.
    * @param permanentId
    *   The [[PermanentId]] of the test.
    * @param tags
    *   List of [[TestDefinition.Tag]] applicable to this test.
    * @param markers
    *   List of [[TestDefinition.Marker]] applicable to this test.
    * @param documentation
    *   The documentation for this test.
    * @param iterations
    *   Number of iterations to run this test.
    */
  final protected class TestBuilder[F[_]: Async](
    val registry: Registry[F],
    val name: TestDefinition.Name,
    val permanentId: PermanentId,
    private val tags: ListBuffer[Tag],
    private val markers: ListBuffer[Marker],
    private var documentation: Option[String] = None,
    private var iterations: TestIterations = TestIterations.One
  ):

    /** Supply documentation for this test.
      *
      * @param docs
      *   The documentation for this test.
      * @return
      *   This builder.
      */
    def document(docs: String): TestBuilder[F] =
      documentation = Some(docs)
      this

    /** Add additional [[Test.Tag]] to this test definition.
      *
      * @param additionalTags
      *   The list of new tags.
      * @return
      *   This builder.
      */
    def tagged(additionalTags: Tag*): TestBuilder[F] =
      val _ = tags.addAll(additionalTags)
      this

    /** Add the [[TestDefinition.Marker.Ignored]] marker to this test
      * definition.
      *
      * @return
      *   This builder.
      */
    def ignored(): TestBuilder[F] =
      val _ = markers.addOne(Marker.Ignored)
      this

    /** Add one or more [[TestDefinition.Marker]] to this test definition.
      *
      * @param additionalMarkers
      *   The list of markers to add.
      * @return
      *   This builder.
      */
    def marked(additionalMarkers: Marker*): TestBuilder[F] =
      val _ = markers.addAll(additionalMarkers)
      this

    /** Set the number of times this test should iterate.
      *
      * @param iters
      *   The number of iterations.
      * @return
      *   This builder.
      */
    def iterate(iters: TestIterations): TestBuilder[F] =
      iterations = iters
      this

    /** Provide an input supplier for this test. Note that each iteration of the
      * test results in the input function being evaluated.
      *
      * @param f
      *   The input function.
      * @return
      *   Builder that supports input.
      */
    def input[Input](f: F[Input]): InputTestBuilder[F, Input] =
      new InputTestBuilder[F, Input](
        registry = registry,
        name = name,
        permanentId = permanentId,
        inputFunction = f,
        tags = tags,
        markers = markers,
        iterations = iterations
      )

    /** Finalize and register this test with a pure unit of work.
      *
      * @param unitOfWork
      *   The function this test will execute.
      */
    def pure(unitOfWork: => Either[TestFailure, Unit]): Unit =
      apply(EitherT.fromEither[F](unitOfWork))

    /** Finalize and register this test with an effectful unit of work.
      *
      * @param unitOfWork
      *   The function this test will execute.
      */
    def effectful(unitOfWork: => F[Either[TestFailure, Unit]]): Unit =
      apply(EitherT(unitOfWork))

    /** Finalize and register this test with an effectful unit of work.
      *
      * @param unitOfWork
      *   The function this test will execute.
      */
    def apply(unitOfWork: => EitherT[F, TestFailure, Unit]): Unit =
      registry.register(
        new TestDefinition[F](
          name = name,
          permanentId = permanentId,
          documentation = documentation,
          tags = tags.distinct.toList,
          markers = markers.distinct.toList,
          iterations = iterations,
          unitOfWork = unitOfWork
        )
      )

  /** Builder to assist with defining tests. This builder is for tests which
    * accept input via some producing function.
    *
    * @param registry
    *   Registry instance internal to a [[TestGroup]] for recording completed
    *   definitions.
    * @param name
    *   The name of the test.
    * @param permanentId
    *   The [[PermanentId]] of the test.
    * @param inputFunction
    *   The function that provides input to this test.
    * @param tags
    *   List of [[TestDefinition.Tag]] applicable to this test.
    * @param markers
    *   List of [[TestDefinition.Marker]] applicable to this test.
    * @param documentation
    *   The documentation for this test.
    * @param iterations
    *   Number of iterations to run this test.
    */
  final protected class InputTestBuilder[F[_]: Async, Input](
    val registry: Registry[F],
    val name: TestDefinition.Name,
    val permanentId: PermanentId,
    val inputFunction: F[Input],
    private val tags: ListBuffer[Tag],
    private val markers: ListBuffer[Marker],
    private var documentation: Option[String] = None,
    private var iterations: TestIterations = TestIterations.One
  ):

    /** Supply documentation for this test.
      *
      * @param docs
      *   The documentation for this test.
      * @return
      *   This builder.
      */
    def document(docs: String): InputTestBuilder[F, Input] =
      documentation = Some(docs)
      this

    /** Add additional [[Test.Tag]] to this test definition.
      *
      * @param additionalTags
      *   The list of new tags.
      * @return
      *   This builder.
      */
    def tagged(additionalTags: Tag*): InputTestBuilder[F, Input] =
      val _ = tags.addAll(additionalTags)
      this

    /** Add the [[TestDefinition.Marker.Ignored]] marker to this test
      * definition.
      *
      * @return
      *   This builder.
      */
    def ignored(): InputTestBuilder[F, Input] =
      val _ = markers.addOne(Marker.Ignored)
      this

    /** Add one or more [[TestDefinition.Marker]] to this test definition.
      *
      * @param additionalMarkers
      *   The list of markers to add.
      * @return
      *   This builder.
      */
    def marked(additionalMarkers: Marker*): InputTestBuilder[F, Input] =
      val _ = markers.addAll(additionalMarkers)
      this

    /** Set the number of times this test should iterate.
      *
      * @param iters
      *   The number of iterations.
      * @return
      *   This builder.
      */
    def iterate(iters: TestIterations): InputTestBuilder[F, Input] =
      iterations = iters
      this

    /** Finalize and register this test with a pure unit of work.
      *
      * @param unitOfWork
      *   The function this test will execute.
      */
    def pure(unitOfWork: Input => Either[TestFailure, Unit]): Unit =
      apply(input => EitherT(Async[F].delay(unitOfWork(input))))

    /** Finalize and register this test with an effectful unit of work.
      *
      * @param unitOfWork
      *   The function this test will execute.
      */
    def effectful(unitOfWork: Input => F[Either[TestFailure, Unit]]): Unit =
      apply(input => EitherT(unitOfWork(input)))

    /** Finalize and register this test with an effectful unit of work.
      *
      * @param unitOfWork
      *   The function this test will execute.
      */
    def apply(unitOfWork: Input => EitherT[F, TestFailure, Unit]): Unit =
      registry.register(
        new TestDefinition[F](
          name = name,
          permanentId = permanentId,
          documentation = documentation,
          tags = tags.distinct.toList,
          markers = markers.distinct.toList,
          iterations = iterations,
          unitOfWork = EitherT.right(inputFunction).flatMap(unitOfWork)
        )
      )

  final protected class Registry[F[_]]:

    val mapping: ConcurrentHashMap[PermanentId, TestDefinition[F]] =
      new ConcurrentHashMap[PermanentId, TestDefinition[F]]

    def register(test: TestDefinition[F]): Unit =
      if mapping.contains(test.permanentId) then
        throw new IllegalArgumentException(
          s"Attempted to register test with duplicate Permanent ID '${test.permanentId.show}'."
        )
      else mapping.put(test.permanentId, test)

    def toList(): List[TestDefinition[F]] = mapping.values().asScala.toList

  end Registry

end TestGroup
